/*
	Evil Bricks, a Breakout clone for SVGAlib
    Copyright (C) 2001 Brian Mastrobuono <gauze@rifug.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>
#include <math.h>
#include <vga.h>
#include <vgagl.h>
#include <vgakeyboard.h>
#include <vgamouse.h>
#ifndef NOSOUND
#include "musicalbeep.h"
#endif

typedef int bool;
#define TRUE 1
#define FALSE 0

#define SLOW 0
#define FAST 1

#define LEFT 0 /* X */
#define RIGHT 1
#define DOWN 0 /* Y */
#define UP 1

#define PADDLETRACK 140


#define MATCHES 0 /* for strcmp matches */

typedef struct Player {
	int score;
	int balls_left;
} Player; 

typedef struct Ball {
	int posX;
	int posY;
	int oldposX;
	int oldposY;
	bool dirX;
	bool dirY;
	bool SPEED;
	float speedX;
} Ball;

typedef struct Paddle {
	int paddlepos;
	int paddlesize;
	int oldpaddlepos;
	int oldpaddlesize;
} Paddle;

const int brickset[8][13] = {	{3,3,3,3,3,3,3,3,3,3,3,3,3},
						  		{2,2,2,2,2,2,2,2,2,2,2,2,2}, 
						 	 	{2,2,2,2,2,2,2,2,2,2,2,2,2},   
						  		{2,2,2,2,2,2,2,2,2,2,2,2,2},   
							  	{2,2,2,2,2,2,2,2,2,2,2,2,2},   
							  	{1,1,1,1,1,1,1,1,1,1,1,1,1},   
						  		{1,1,1,1,1,1,1,1,1,1,1,1,1},   
						  		{1,1,1,1,1,1,1,1,1,1,1,1,1}	};

typedef struct Bricks {
	int bricksleft;
	int bricks[8][13]; /* memcpy brickset[][] above into it in initfield() */
} Bricks;

/* global flags */
bool USE_MOUSE = TRUE;
bool GAMEOVER = FALSE; 
bool BALLINPLAY = TRUE;
bool LAST_HIT_NOT_BRICK = TRUE; /* for bounces off the top ... */
bool HIT = FALSE;
bool FLIP = FALSE;
bool NEWGAME = TRUE;
bool SOUND = TRUE;

/* svgalib color chart 0-15 */
enum Colors { BLACK, BLUE, GREEN, CYAN, RED, MAGENTA, BROWN, GREY, DARK_GREY,
			  LIGHT_BLUE, LIGHT_GREEN, LIGHT_CYAN, LIGHT_RED, LIGHT_MAGENTA,
			  YELLOW, WHITE };

/* functions */

void help(void);
void newgame(Player *, Paddle *, Bricks *);
int direction(int);
void initfield(Bricks *);
bool get_keys(Ball *, Paddle *, Bricks *);
bool get_mouse(Ball *, Paddle *, Bricks *);
void draw_bricks(Bricks *);
void draw_paddle(Paddle *);
void draw_ball(Ball *,int);
void action(Player *, Ball *, Paddle *, Bricks *);
void print_score_balls(Player *);
void serve_ball(Ball *, Bricks *);
bool timer(int);
void ppmsnapshot(void);
void hiscore(void);

int my_balls = 0; /* need a global for this */

int main(int argc, char *argv[])
{
	int counter = 0;
	Player *player1 = malloc(sizeof *player1 );
	Paddle *paddle = malloc(sizeof *paddle );
	Ball *ball = malloc(sizeof *ball );
	Bricks *bricks = malloc(sizeof *bricks );
	int opt_index = 0, c;
	static struct option long_options[] = { /* for getopt_long() */
	{"nm",0,0,'m'},
	{"nomouse",0,0,'m'},
	{"ns",0,0,'s'},
	{"nosound",0,0,'s'},
	{"balls",1,0,'b'},
	{0,0,0,0}
	};

	if ( !player1 || !paddle || !ball || !bricks )
	{
	printf("Out of Memory!\n");
	exit(-1);
	} 

		
	srand(time(0)); // for ball placement on serve


	/* svgalib init stuff */
	vga_init();
	vga_setmode(G320x200x256);
	vga_setcolor(WHITE);
	gl_setcontextvga(G320x200x256);
	gl_enableclipping();
	gl_setfont(8, 8, gl_font8x8);
	gl_setwritemode(FONT_COMPRESSED + WRITEMODE_OVERWRITE);
	gl_setfontcolors(0,WHITE);

	if (keyboard_init()) 
	{
		printf("Error: INIT_KEYB!\n");
		exit(-1);
	}
	
	if (argc > 1)
		for(;;)
		{	
			c = (getopt_long_only(argc, argv,"b:", long_options, &opt_index));
			if (c == -1)
				break;
			switch(c)
			{
				case 's': SOUND = FALSE;
					break;
				case 'm': USE_MOUSE = FALSE;
					break;
				case 'b': if (optarg)
							my_balls = atoi(optarg); 
					 	  else
							help();
					break;
				default: help();
					break; 
			}
		}

	if (USE_MOUSE)
	{
		vga_setmousesupport(1);
		mouse_setxrange(0, 120);
		mouse_setwrap(MOUSE_NOWRAP);
	}


	/* end of svgalib init stuff */
	newgame(player1, paddle, bricks);	

	while(1) { /* 'master' loop where EVERYTHING happens */
		draw_bricks(bricks); 
		draw_paddle(paddle); 
		serve_ball(ball, bricks);
		while (!GAMEOVER)
		{
			counter++;
		
			if ( BALLINPLAY && timer(counter) )
			{
				draw_ball(ball, BLACK);
				action(player1, ball, paddle, bricks);
				draw_ball(ball, WHITE);
				usleep(1/(ball->SPEED+1));
				counter = 0;
			}
			if (USE_MOUSE)
			{
				if ( get_mouse(ball, paddle, bricks) )	
					draw_paddle(paddle);
			} 
			if ( get_keys(ball, paddle, bricks) )
				draw_paddle(paddle); 
	
			if ((ball->posY<66 && ball->posX>20) && !(counter%19) )/*maybe 39?*/
				draw_bricks(bricks); 

		}

		gl_printf(115, 75, "GAME OVER");
		sleep(2);
		printf("\nYour Score: %d\n", player1->score);
		gl_printf(50,190,"Play another game? (y/n)");
	
		for(;;)
		{
			keyboard_update();
			if ( keyboard_keypressed(SCANCODE_N) ) 
				{ NEWGAME=FALSE; break; }
			if ( keyboard_keypressed(SCANCODE_Y) ) 
				{ newgame(player1,paddle,bricks); break; }
		}
		gl_printf(50,190,"                        ");
		gl_printf(115,75, "         ");

		if (!NEWGAME)	
			break;
	} /* end outside while(1) loop */

	if (USE_MOUSE)
		mouse_close();
	
	keyboard_close(); 
	free(paddle);
	free(player1);
	free(ball);
	free(bricks);

	vga_setmode(TEXT);

	exit(0);
} /* end of main() */

/* start of function definitions */

void newgame(Player *player1, Paddle *paddle, Bricks *bricks)
{
	paddle->paddlepos = paddle->oldpaddlepos = 0;
	paddle->paddlesize = paddle->oldpaddlesize = 5;
	player1->score = 0;
	if (my_balls > 0)
		player1->balls_left = my_balls; /* comical! */
	else
		player1->balls_left = 3;
	NEWGAME=TRUE;
	GAMEOVER=FALSE;
	vga_clear();
	print_score_balls(player1);
	initfield(bricks);
}

void initfield(Bricks *bricks)
{
	/* the walls */
	gl_fillbox(89,  0, 135, 5, WHITE);
	gl_fillbox(85,  0, 5, 150, WHITE);
	gl_fillbox(221, 0, 5, 150, WHITE);

	bricks->bricksleft = 104;
	memcpy(bricks->bricks, brickset, sizeof(bricks->bricks));
//	memcpy(bricks->bricks, brickset, sizeof(Bricks));
	return;			
}

void draw_bricks(Bricks *bricks)
{
	int x, y, color;

	for ( y = 0 ; y < 8 ; y++ )
		for ( x = 0 ; x < 13 ; x++ )
		{
			color = (LIGHT_GREEN); /* 1 point bricks */
			if ( bricks->bricks[y][x] == 2 ) color = (RED);
			if ( bricks->bricks[y][x] == 3 ) color = (BLUE);
			if ( bricks->bricks[y][x] == 0 ) color = (BLACK);

			gl_fillbox(x*10+90+1, y*5+24+1, 9, 4, color);
			HIT = FALSE;
		}

	return;
}

void draw_paddle(Paddle *paddle)
{
	gl_fillbox(paddle->oldpaddlepos+90, PADDLETRACK, 10, 5, BLACK );
	gl_fillbox(paddle->paddlepos+90, PADDLETRACK, 10, 5, BLUE );
	paddle->oldpaddlepos = paddle->paddlepos;
	paddle->oldpaddlesize = paddle->paddlesize;
	return;
}


void print_score_balls(Player *player1)
{
	gl_printf(90,155,"                ");
	gl_printf(90,155,"%d            %d", player1->balls_left, player1->score);
	return;
}

void silent_pause(void)
{
	gl_printf(0,0,"      ");
	for (;;)
	{
		if (USE_MOUSE)
		{
			if ( mouse_update() )
				if (mouse_getbutton() & MOUSE_LEFTBUTTON )
					break;
		}

    	if ( keyboard_update() )
            if (keyboard_keypressed(SCANCODE_P) || 
				keyboard_keypressed(SCANCODE_SPACE) ) 
				break;
		
	}
	return;
}

void paused(void)
{
	gl_printf(0,0,"PAUSED");
	for(;;)
	{
    	if ( keyboard_update() )
       		if (keyboard_keypressed(SCANCODE_P))
				break;
	}
	gl_printf(0,0,"      ");
	return;
}

bool get_keys(Ball *ball, Paddle *paddle, Bricks *bricks)
{
	if ( keyboard_update() )
	{
		if ( keyboard_keypressed(SCANCODE_CURSORBLOCKLEFT) )
		{
			  if ( paddle->paddlepos == 0 ) return(FALSE); /* cycle saver */
			  if ( paddle->paddlepos > 0 ) paddle->paddlepos-=4;
			  if ( paddle->paddlepos == 1 ) paddle->paddlepos = 0;/*see below*/
			  return(TRUE);
		}
		if ( keyboard_keypressed(SCANCODE_CURSORBLOCKRIGHT) )
		{
			  if ( paddle->paddlepos == 121 ) return(FALSE);/* bail if touching*/
			  if ( paddle->paddlepos < 120 ) paddle->paddlepos+=4;
			  if ( paddle->paddlepos == 120 ) paddle->paddlepos++; /* kludge :*/
			  return(TRUE);
		}
		if ( keyboard_keypressed(SCANCODE_P) )
		{
			paused();
			return(TRUE);
		}

		if (!BALLINPLAY && keyboard_keypressed(SCANCODE_SPACE) ) 
		{
			serve_ball(ball, bricks);
			return(TRUE);
		}
		
		if ( keyboard_keypressed(SCANCODE_ESCAPE) )
			GAMEOVER = TRUE;
		
		if ( keyboard_keypressed(SCANCODE_T) ) /*'take' snapshot */
			ppmsnapshot();

	}
	return(FALSE);
}

bool get_mouse(Ball *ball, Paddle *paddle, Bricks *bricks)
{
	mouse_update();
	
	if ( !BALLINPLAY && (mouse_getbutton() & MOUSE_LEFTBUTTON) )
	{
		serve_ball(ball, bricks);
		return(TRUE);
	}
	
	if ( mouse_getx() == paddle->paddlepos && !BALLINPLAY)
		return(FALSE);

	
	if (mouse_getx() < 120)
	{
		paddle->paddlepos = mouse_getx();
		return(TRUE);
	}
	else if (paddle->paddlepos != 121)
	{
		paddle->paddlepos = 121;
		return(TRUE);
	}

	return(FALSE);
}

void serve_ball(Ball *ball, Bricks *bricks)
{
	if (bricks->bricksleft < 1)
		initfield(bricks);	
    /* kill old ball->pos */
	draw_ball(ball,BLACK);

	ball->SPEED = SLOW;
	ball->posX = (int)(120.0*rand()/(RAND_MAX+1.0) + 95); 
	ball->posY = 64; /* always start right under bricks */
	ball->speedX = (float)(3.0*rand()/(RAND_MAX+1.0));
	if (ball->speedX < 1.0) ball->speedX = 1.0;
	ball->dirX = ( ( ( 10.0*rand()/(RAND_MAX+1.0) ) > 5 ) ? RIGHT : LEFT);
	ball->dirY = DOWN;
	silent_pause();
	BALLINPLAY = TRUE;
	return;
}

void draw_ball(Ball *ball, int color)
{
	gl_fillbox(ball->posX, ball->posY, 2, 2, color);
	return;
}

void action(Player *player1, Ball *ball, Paddle *paddle, Bricks *bricks)
{	
	int i, j;
#ifndef NOSOUND
	beep_parms_t thebeep;
#endif
	
	ball->oldposX = ball->posX;
	ball->oldposY = ball->posY;	
	
	/* ball is moving ... */	
	if (ball->dirX == RIGHT)
		ball->posX += (int)ball->speedX; 
	if (ball->dirX == LEFT)
		ball->posX -= (int)ball->speedX;
	if (ball->dirY == DOWN)
		ball->posY += 1;
	if (ball->dirY == UP)
		ball->posY -= 1;

	if (ball->posX < 90 && ball->dirX == LEFT)  /* left wall bounce */
	{
		ball->dirX = RIGHT;
		ball->posX = 90;
#ifndef NOSOUND
		if (SOUND)
		{
			thebeep.note = NOTE_G;
			thebeep.octave = 1;
			thebeep.length = EIGHTH_NOTE;
			thebeep.tempo = 320;
			play_beep(thebeep);
		}
#endif
	}
	if (ball->posX > 219 && ball->dirX == RIGHT) /* right wall bounce */
	{
		ball->dirX = LEFT;
		ball->posX = 219;
#ifndef NOSOUND
		if (SOUND)
		{
		thebeep.note  = NOTE_E;
		thebeep.octave = 1;
		thebeep.length = EIGHTH_NOTE;
		thebeep.tempo = 320;
		play_beep(thebeep);
		}
#endif
	}
	if ( ball->posY < 5 && ball->dirY == UP )  /* top of screen */
	{
		ball->dirY = DOWN;
		ball->posY = 5;
		ball->SPEED = FAST;
#ifndef NOSOUND
		if (SOUND)
		{
		thebeep.note = NOTE_B;
		thebeep.octave = 2;
		thebeep.length = EIGHTH_NOTE;
		thebeep.tempo = 320;
		play_beep(thebeep);
		}
#endif
		LAST_HIT_NOT_BRICK = TRUE;
	}
	if ( ball->posY > 150 )  /* oops in the pit ... next ball */
	{
		player1->balls_left--; 
		BALLINPLAY = FALSE;
		draw_paddle(paddle);
#ifndef NOSOUND
		if (SOUND)
		{
		thebeep.note = NOTE_A;
		thebeep.octave = 0.5;
		thebeep.length = HALF_NOTE;
		thebeep.tempo = 120;
		play_beep(thebeep);
		}
#endif
		if ( player1->balls_left == 0 )
			GAMEOVER=TRUE;
		print_score_balls(player1);
	}

	/* touching the paddle? */
	if ( ball->posY > 137 )
	{
	for ( i = 0 ; i < paddle->paddlesize*2 ; i++ )
		for ( j = 0 ; j < 4 ; j++ ) /* arbitrary bottom */
			if ( ball->posY == 140+j && ball->posX == (paddle->paddlepos+90+i) )
			{
				if (bricks->bricksleft < 1) /* cleared a screen! */
				{            				/* so redraw the screen dummy! */
					initfield(bricks);
					draw_bricks(bricks);
				}
				ball->dirY = UP;
				ball->posY = 138;
#ifndef NOSOUND
				if (SOUND)
				{
				thebeep.note = NOTE_A;
				thebeep.octave = 2;
				thebeep.length = EIGHTH_NOTE;
				thebeep.tempo = 320;
				play_beep(thebeep);
				}
#endif
				draw_paddle(paddle);
				
				/* handle ball trajectory via X axis speed ... */
				/* I fucking hate how this works now argh! */
				if ( ball->speedX == 0.0 ) 
				{
					if ( i == 0 ) { ball->speedX += 2.0; ball->dirX = LEFT; }
					if ( i == 1 ) { ball->speedX += 1.5; ball->dirX = LEFT; }
					if ( i == 2 ) { ball->speedX += 1.0; ball->dirX = LEFT; }
/*
					if ( i == 3 ) { ball->speedX += 0.5; ball->dirX = LEFT; }
					if ( i == 4 ) { ball->speedX += .25; ball->dirX = LEFT; }
					if ( i == 5 ) { ball->speedX += .25; ball->dirX = RIGHT; }
					if ( i == 6 ) { ball->speedX += 0.5; ball->dirX = RIGHT; }
*/
					if ( i == 3 ) { ball->speedX++; ball->dirX = LEFT; }
                    if ( i == 4 ) { ball->speedX++; ball->dirX = LEFT; }
                    if ( i == 5 ) { ball->speedX++; ball->dirX = RIGHT; }
                    if ( i == 6 ) { ball->speedX++; ball->dirX = RIGHT; }
	
					if ( i == 7 ) { ball->speedX += 1.0; ball->dirX = RIGHT; }
					if ( i == 8 ) { ball->speedX += 1.5; ball->dirX = RIGHT; }
					if ( i == 9 ) { ball->speedX += 2.0; ball->dirX = RIGHT; }
				}
				else if ( ball->dirX == RIGHT )
				{		
					if ( i == 0 ) FLIP = TRUE;
					if ( i == 1 ) ball->speedX-=1.5;
					if ( i == 2 ) ball->speedX--;
					if ( i == 3 ) ball->speedX-=.5;
					if ( i == 4 ) ball->speedX-=.25; 
					if ( i == 5 ) ball->speedX+=.25; 
					if ( i == 6 ) ball->speedX+=.5;
					if ( i == 7 ) ball->speedX++;
					if ( i == 8 ) ball->speedX+=1.5;
					if ( i == 9 ) ball->speedX+=2;
				}
				else if ( ball->dirX == LEFT )
				{
					if ( i == 0 ) ball->speedX+=2;
					if ( i == 1 ) ball->speedX+=1.5;
					if ( i == 2 ) ball->speedX++;
					if ( i == 3 ) ball->speedX+=.5;
	  				if ( i == 4 ) ball->speedX+=.25; 
					if ( i == 5 ) ball->speedX-=.25; 
					if ( i == 6 ) ball->speedX-=.5;
					if ( i == 7 ) ball->speedX--;
					if ( i == 8 ) ball->speedX-=1.5;
					if ( i == 9 ) FLIP = TRUE;
				}
				if (ball->speedX < 0.0 || FLIP) /*flip direction and speed */
				{
					ball->dirX = ball->dirX?LEFT:RIGHT;
					ball->speedX = fabsf(ball->speedX); 
					FLIP = FALSE;
				}
				if ( ball->speedX < 1.0 ) 
					ball->speedX++;/* kludge */

				if (ball->speedX > 3.0) ball->speedX = 3.0; /*ball2fast?*/
					LAST_HIT_NOT_BRICK = TRUE;
			}
	}
	/* hitting a brick? */
	if ((ball->posY > 23 && ball->posY < 65) && LAST_HIT_NOT_BRICK)
	{
		float currentY, currentX;
		currentY=(ball->posY-22)/5;
		currentX=(ball->posX-89)/10;

		if (currentX < 0) currentX = 0; /* safety for math 'errors' */
		if (currentY < 0) currentY = 0; 
		if (currentY > 7.0) currentY = 7.0; 
		if (currentX > 12.0) currentX = 12.0;

		if (ball->dirY == UP || LAST_HIT_NOT_BRICK)
		{
			if ((bricks->bricks)[(int)currentY][(int)currentX] > 0 
				 && LAST_HIT_NOT_BRICK )
			{
				ball->dirY=(ball->dirY?DOWN:UP);/* switch direction */
#ifndef NOSOUND
				if (SOUND)
				{
				thebeep.note = NOTE_C;
				thebeep.octave = 2;
				thebeep.length = EIGHTH_NOTE;
				thebeep.tempo = 320;
               	play_beep(thebeep);
				}
#endif
				bricks->bricksleft--;
				player1->score+=(bricks->bricks)[(int)currentY][(int)currentX];
				(bricks->bricks)[(int)currentY][(int)currentX]=0;
				print_score_balls(player1);
				HIT = TRUE;
				draw_bricks(bricks);
				LAST_HIT_NOT_BRICK = FALSE;
			}
		}
	} 
	
	return;
}



bool timer(int count)
{
	if ( count >= 40 )
		return(TRUE);

	return(FALSE);
}

void help(void)
{
	printf("\nEvil Bricks - copyright 2001 Brian Mastrobuono <gauze@rifug.org>\n");
	printf("GNU GPL - no warranty, blah, blah, ...\n\n");
	printf("-b # or --balls # \t number of balls to start game with\n");
	printf("--nomouse or -nm to turn off mouse support\n");
#ifndef NOSOUND
	printf("--nosound or -ns to turn off sound support\n");
#endif
	printf("\nany other arg besides the mouse ones above: this message\n");
	exit(0);
}


void ppmsnapshot(void)
{
    FILE *file;
    int x, y, width = 320, height = 200; 
	static int num = 0; 
	char filename[20]; /* hope no one takes > 9999999 snaps */
	int *r = malloc(sizeof(*r));
	int *g = malloc(sizeof(*g));
	int *b = malloc(sizeof(*b)); 
	if ( !r || !g || !b )
	{
		puts("Out of Memory!\n");
		exit(-1);
	} 

	sprintf(filename, "%s%d%s", "snapshot", ++num, ".ppm"); 
	
    file = fopen(filename, "w");
	if (file == NULL)
		return;
    fprintf(file, "P3\n%d %d\n63\n", width, height);

    for (y = 0; y < height; y++) 
	{
		if (y > 0) fprintf(file,"\n");
			for (x = 0; x < width; x++) 
			{
    			gl_getpalettecolor(gl_getpixel(x, y) , r, g, b);
				fprintf(file, "%d %d %d ", *r, *g, *b);
			}
    }
	fclose(file); 
	printf("\nSnapshot taken\n"); 
	free(r);
	free(g);
	free(b); 
	return; 
}

void hiscore(void) { /* placeholder */ }
