/* musicalbeep.h */
/* derived from beep-1.2.1 (see below) */
/* contains about a dozen lines from beep-1.2.1 */
/* copyright Brian Mastrobuono, 2001 */
/* add (possibly musical) console beeps to your C programs ( I have used it 
	in tetris and breakout clones so far, for example) */

/*  beep - just what it sounds like, makes the console beep - but with
 * precision control.  See the man page for details.
 *
 * Try beep -h for command line args
 *
 * This code is copyright (C) Johnathan Nightingale, 2000.
 *
 * This code may distributed only under the terms of the GNU Public License
 * which can be found at http://www.gnu.org/copyleft or in the file COPYING
 * supplied with this code.
 *
 * This code is not distributed with warranties of any kind, including implied
 * warranties of merchantability or fitness for a particular use or ability to
 * breed pandas in captivity, it just can't be done.
 *
 * Bug me, I like it:  http://johnath.com/  or johnath@johnath.com
 */

/* for the library version here try: gauze@rifug.org */
/* http://www.rifug.org/~gauze/stuff/ */


#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/kd.h>

#ifndef CLOCK_TICK_RATE
#define CLOCK_TICK_RATE 1193180
#endif

#define REST 0

#define NOTE_A 220.0
#define NOTE_A_SHARP 233.0818808 
#define NOTE_B_FLAT NOTE_A_SHARP
#define NOTE_B 246.9416506
#define NOTE_C 261.6255653
#define NOTE_C_SHARP 277.182631
#define NOTE_D_FLAT NOTE_C_SHARP
#define NOTE_D 293.6647679
#define NOTE_D_SHARP 311.1269837
#define NOTE_E 329.6275569
#define NOTE_F 349.2282314
#define NOTE_F_SHARP 369.9944227
#define NOTE_G_FLAT NOTE_F_SHARP
#define NOTE_G 391.995436
#define NOTE_G_SHARP 415.3046976
#define NOTE_A_FLAT NOTE_G_SHARP



#define WHOLE_NOTE 64
#define HALF_NOTE 32 
#define QUARTER_NOTE 16
#define EIGHTH_NOTE 8
#define SIXTEENTH_NOTE 4
#define THIRTYSECOND_NOTE 2
#define SIXTYFOURTH_NOTE 1
 

typedef struct beep_parms_t 
{
	float note;   /* which note of 12 tone scale */
	float octave;  /* relative octave note is in */
	int length;  /* Note length value (WHOLE_NOTE, HALF, etc)       */
	int tempo; /* tempo, 4/4, whole note  */
} beep_parms_t;

int console_fd = -1;

void play_beep(beep_parms_t parms) 
{
/*	console_fd = open(ttyname(STDOUT_FILENO), O_WRONLY); */
/* 	line below is from beep.c ... use it if you wish and comment above line */
	console_fd = open("/dev/console", O_WRONLY); 
	ioctl(console_fd, KIOCSOUND, (int)(CLOCK_TICK_RATE/(parms.note*parms.octave))); 
    usleep((1000000/parms.tempo*4)*(parms.length));/* wait... tempo*4 4/4   */
    ioctl(console_fd, KIOCSOUND, 0);    /* stop beep  */
	close(console_fd); 
}
